/* Mini systeme de particules
 * Auteur : Fares BELHADJ (Lire COPYING)
 * Email  : amsi@ai.univ-paris8.fr
 * Date   : 03/01/2008
 */
#include <math.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <stdlib.h>
#include <assert.h>
#include "particules.h"

particule_t * particules = NULL;
vector3f_t G = {0, -9.8, 0};

static int    n_particules      = 0;
static GLuint gl_unit_particule = 0;

static void drawParticule       (particule_t * p);
static void moveParticule       (particule_t * p, float dt);
static void frottements         (particule_t * p, float k);
static void ricochet            (particule_t * p, float dt);
static void makeGLUnitParticule (void);

extern void initParticules(int n) {
  float s;
  int i, j;
  static int firstTime = 1;
  if(firstTime) {
    makeGLUnitParticule();
    atexit(freeParticules);
    firstTime = 0;
  }
  freeParticules();
  particules = malloc((n_particules = n) * sizeof * particules);
  assert(particules);
  for(i = 0; i < n_particules; i++) {
    particules[i].r   = 0.02 + 0.06 * u_rand();

    particules[i].p.x = su_rand() / 2.0;
    particules[i].p.y = su_rand() / 2.0;
    particules[i].p.z = su_rand() / 2.0;
    
    particules[i].v.x = su_rand();
    particules[i].v.y = 4.0 * u_rand();
    particules[i].v.z = su_rand();
    
    particules[i].a   = G;

    do {
      s = 0;
      for(j = 0; j < 3; j++)
	s += (particules[i].ambient_color[j] = MAX(0, su_rand()));
    } while(s < 0.2);
    do {
      s = 0;
      for(j = 0; j < 3; j++)
	s += (particules[i].diffuse_color[j] = MAX(0, su_rand()));
    } while(s < 0.6);
    particules[i].ambient_color[3] = particules[i].diffuse_color[3] = 1.0;
  }
}

extern void drawParticules(void) {
  int i;
  assert(particules);
  for(i = 0; i < n_particules; i++)
    drawParticule(&particules[i]);
}

extern void moveParticules(float dt) {
  int i;
  assert(particules);
  for(i = 0; i < n_particules; i++)
    moveParticule(&particules[i], dt);
}

extern void freeParticules(void) {
  if(particules)
    free(particules);
  particules = NULL;
  n_particules = 0;
}

static void drawParticule(particule_t * p) {
  glPushMatrix();
  glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, p->ambient_color);
  glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, p->diffuse_color);
  glTranslatef(p->p.x, p->p.y, p->p.z);
  glScalef(p->r, p->r, p->r);
  glCallList(gl_unit_particule);
  glPopMatrix();
}

static void moveParticule(particule_t * p, float dt) {
  p->a = G;
  p->v.x += p->a.x * dt;
  p->v.y += p->a.y * dt;
  p->v.z += p->a.z * dt;
  frottements(p, 0.0005);
  p->p.x += p->v.x * dt;
  p->p.y += p->v.y * dt;
  p->p.z += p->v.z * dt;
  ricochet(p, dt);
}

static void frottements(particule_t * p, float k) {
  float v, s;
  v = fabs(p->v.x);
  s = p->v.x < 0.0 ? -1 : 1;
  v = v - (k + k * su_rand()) * v * v;
  p->v.x = s * MAX(v, 0);

  v = fabs(p->v.y);
  s = p->v.y < 0.0 ? -1 : 1;
  v = v - (k + k * su_rand()) * v * v;
  p->v.y = s * MAX(v, 0);

  v = fabs(p->v.z);
  s = p->v.z < 0.0 ? -1 : 1;
  v = v - (k + k * su_rand()) * v * v;
  p->v.z = s * MAX(v, 0);
}

static void ricochet(particule_t * p, float dt) {
  if(p->p.x < -1) {
    p->p.x = -1;
    if(p->v.x < 0.0) p->v.x = - p->v.x;
    frottements(p, 0.005);
  } else if(p->p.x > 1) {
    p->p.x = 1;
    if(p->v.x > 0.0) p->v.x = - p->v.x;
    frottements(p, 0.005);
  }
  if(p->p.y < -1) {
    p->p.y = -1;
    if(p->v.y < 0.0) p->v.y = - p->v.y;
    frottements(p, 0.005);
  } else if(p->p.y > 1) {
    p->p.y = 1;
    if(p->v.y > 0.0) p->v.y = - p->v.y;
    frottements(p, 0.005);
  }
  if(p->p.z < -1) {
    p->p.z = -1;
    if(p->v.z < 0.0) p->v.z = - p->v.z;
    frottements(p, 0.005);
  } else if(p->p.z > 1) {
    p->p.z = 1;
    if(p->v.z > 0.0) p->v.z = - p->v.z;
    frottements(p, 0.005);
  }
}

static void makeGLUnitParticule(void) {
  GLUquadricObj * q;
  q = gluNewQuadric();
  assert(q);
  gluQuadricNormals(q, GLU_SMOOTH); /* smooth est par defaut */
  gluQuadricTexture(q, GL_FALSE); /* false est par defaut */

  gl_unit_particule = glGenLists(1);
  glNewList(gl_unit_particule, GL_COMPILE);
  gluSphere(q, 1.0, 6, 5);
  glEndList();

  gluDeleteQuadric(q);
}
