/* Mini systeme de particules dans un cube.
 * Auteur : Fares BELHADJ (Lire COPYING)
 * Email  : amsi@ai.univ-paris8.fr
 * Date   : 03/01/2008
 */

#include <math.h>
#include <time.h>
#include <stdio.h>
#include <stdlib.h>
#include <SDL.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include "particules.h"

static SDL_Surface * sdlInit(int w, int h);
static void          sdlInitGL(SDL_Surface * s);
static void          sdlResizeGL(SDL_Surface * s);
static void          sdlLoop(SDL_Surface * s);
static int           sdlManageEvents(SDL_Surface * s);
static void          sdlDraw(SDL_Surface * s);
static inline void   sdlFPS(void);
static void cube(void);


static int pause = 0;

/* pour le clavier */
static int k_haut = 0, k_bas = 0, k_droite = 0, k_gauche = 0;

/* la perspective et les angles de rotation */
static float near = 1, far = 15, rx = 0, rz = 0;

/* un tout petit main */
int main(int argc, char ** argv) {
  SDL_Surface * s;
  srand(time(NULL));
  s = sdlInit(640, 480);
  sdlLoop(s);
  return 0;
}

/* sdlInit : initialiser SDL pour OpenGL */
static SDL_Surface * sdlInit(int w, int h) {
  SDL_Surface * s;
  const SDL_VideoInfo * vi = NULL;
  if ( SDL_Init(SDL_INIT_VIDEO) < 0 ) {
    fprintf(stderr, "Impossible d'initialiser SDL: %s\n", SDL_GetError());
    exit(1);
  }

  if( (vi = SDL_GetVideoInfo()) == 0 ) {
    fprintf(stderr, "Echec de SDL_GetVideoInfo(): %s\n", SDL_GetError());
    exit(1);
  }
  SDL_GL_SetAttribute( SDL_GL_RED_SIZE, 5 );
  SDL_GL_SetAttribute( SDL_GL_GREEN_SIZE, 5 );
  SDL_GL_SetAttribute( SDL_GL_BLUE_SIZE, 5 );
  SDL_GL_SetAttribute( SDL_GL_DEPTH_SIZE, 16 );
  SDL_GL_SetAttribute( SDL_GL_DOUBLEBUFFER, 1 );
  if ( (s = SDL_SetVideoMode(w, h, vi->vfmt->BitsPerPixel,
			     SDL_OPENGL | SDL_RESIZABLE)) == NULL ) {
    fprintf(stderr, "Impossible d'ouvrir le mode video : %s\n", SDL_GetError());
    exit(1);
  }
  atexit(SDL_Quit);
  sdlInitGL(s);
  return s;
}

/* 
 * sdlInitGL : configurer les options GL et lancer l'init des
 * particules.
 */
static void sdlInitGL(SDL_Surface * s) {
  glShadeModel(GL_SMOOTH);
  glDepthFunc(GL_LEQUAL);
  glFrontFace(GL_CCW);
  glCullFace(GL_BACK);
  glClearColor(0, 0, 0, 0);
  glClearDepth(1.0);
  glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
  glEnable(GL_CULL_FACE);
  glEnable(GL_DEPTH_TEST);

  glPolygonMode(GL_FRONT, GL_FILL);
  glPolygonMode(GL_BACK, GL_LINE);

  glEnable(GL_LIGHTING);
  glEnable(GL_LIGHT0);
  
  glDisable(GL_TEXTURE_2D);

  initParticules(200);

  sdlResizeGL(s);
}

/* 
 * CallBack : en cas de resize modifier la matrice de projection en
 * fonction du rapport w/h
 */
static void sdlResizeGL(SDL_Surface * s) {
  glViewport(0, 0, s->w, s->h);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluPerspective(45.0, s->w / (float)s->h, near, far);
  glMatrixMode( GL_MODELVIEW );
  glLoadIdentity( );
}

/* 
 * loop principale : sdlManageEvents renvoie vrai si la fenetre est
 * inactive => tantque inactive dors une milliseconde.
 * sinon dessiner puis afficher le framerate
 *
 * regarder l'etat des touches pressees et modifier les angles de
 * roation et calculer les coordonnees de G par rotations inverses.
 */
static void sdlLoop(SDL_Surface * s) {
  int i;
  Uint32 dt;
  Uint32 t0 = 0;
  float matrice[16], tmp[4];

  for(;;) {
    dt = SDL_GetTicks() - t0;
    t0 += dt;
    while(sdlManageEvents(s)) SDL_Delay(1);

    if(k_droite) rz -= 180 * dt / 1000.0;
    if(k_gauche) rz += 180 * dt / 1000.0;

    if(k_haut)   rx -= 180 * dt / 1000.0;
    if(k_bas )   rx += 180 * dt / 1000.0;
   
    glPushMatrix();
    glLoadIdentity();

    glRotatef(-rz, 0, sin(rx * M_PI / 180.0), cos(rx * M_PI / 180.0));
    glRotatef(-rx, 1, 0, 0);

    glGetFloatv( GL_MODELVIEW_MATRIX, matrice );
    glPopMatrix();

    for (i = 0; i < 4; i++)
      tmp[i] = -9.8 * matrice[i + 4] + matrice[i + 12];
    G.x = tmp[0] / tmp[3];
    G.y = tmp[1] / tmp[3];
    G.z = tmp[2] / tmp[3];
 
    if(!pause)
      moveParticules(dt / 1000.0);

    sdlDraw(s);
    sdlFPS();
  }
}

/* 
 * sdlManageEvents gere les evenements ; aussi renvoie vrai si la
 * fenetre est inactive, sinon faux.
 * Les evenements sont : 
 *
 * Les fleches pour modifier les angles
 *
 * escape, q, Q : quitter
 *
 * r : reinitialiser les particules
 *
 * p : pause
 *
 * m : changer de filtrage pour la texture
 *
 * w : activer / desactiver le mode wireframe
 *
 * l : activer / desactiver l'eclairage
 *
 * f : activer / desactiver le dgrad
 *
 * F1 : fullscreen On / Off
 * 
 * redimensionnement de fenetre : appeler sdlResizeGL
 *
 */
static int sdlManageEvents(SDL_Surface * s) {
  GLint v[2];
  static int active = 1;
  SDL_Event event;
  while(SDL_PollEvent(&event)) 
    switch (event.type) {
    case SDL_KEYDOWN:
      switch(event.key.keysym.sym) {
      case 'r':
	initParticules(200);
	break;
      case ' ':
      case 'p':
	pause = !pause;
	break;
      case 'w':
	glGetIntegerv(GL_POLYGON_MODE, v);
	if(v[0] == GL_FILL)
	  glPolygonMode(GL_FRONT, GL_LINE);
	else
	  glPolygonMode(GL_FRONT, GL_FILL);
	break;
      case 'f':
	glGetIntegerv(GL_SHADE_MODEL, v);
	if(v[0] == GL_FLAT)
	  glShadeModel(GL_SMOOTH);
	else
	  glShadeModel(GL_FLAT);
	break;
      case 'l':
	if(glIsEnabled(GL_LIGHTING)) {
	  glDisable(GL_LIGHTING);
	  glDisable(GL_LIGHT0);
	} else {
	  glEnable(GL_LIGHTING);
	  glEnable(GL_LIGHT0);
	}
	break;
      case SDLK_UP:
	k_haut = 1;
	break;
      case SDLK_DOWN:
	k_bas = 1;
	break;
      case SDLK_LEFT:
	k_gauche = 1;
	break;
      case SDLK_RIGHT:
	k_droite = 1;
	break;
      case SDLK_ESCAPE:
      case 'q':
	exit(0);
      case SDLK_F1:
	SDL_WM_ToggleFullScreen(s);
	break;
      default:
	fprintf(stderr, "La touche %s a ete pressee\n",
		SDL_GetKeyName(event.key.keysym.sym));
	break;
      }
      break;
    case SDL_KEYUP:
      switch(event.key.keysym.sym) {
      case SDLK_UP:
	k_haut = 0;
	break;
      case SDLK_DOWN:
	k_bas = 0;
	break;
      case SDLK_LEFT:
	k_gauche = 0;
	break;
      case SDLK_RIGHT:
	k_droite = 0;
	break;
      default:
	break;
      }
      break;
    case SDL_ACTIVEEVENT:
      if(event.active.state & SDL_APPACTIVE)
	active = event.active.gain;
      break;
    case SDL_VIDEORESIZE:
      if( (s = SDL_SetVideoMode(event.resize.w, event.resize.h, 
				s->format->BitsPerPixel, s->flags)) == NULL ) {
	fprintf(stderr, "Echec de SDL_SetVideoMode apres le resize : %s\n", 
		SDL_GetError());
	exit(1);
      }
      sdlResizeGL(s);
      break;
    case SDL_QUIT:
      exit(0);
    }
  return !active;
}

/*
 * Affichage : combinaison (ca n'est pas une composee) de rotations.
 */
static void sdlDraw(SDL_Surface * s) {
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  glLoadIdentity();
  
  gluLookAt(0, 0, 4.5, 0, 0, 0, 0, 1, 0);

  glPushMatrix();
  glRotatef(rx, 1, 0, 0);
  glRotatef(rz, 0, sin(rx * M_PI / 180.0), cos(rx * M_PI / 180.0));
  cube();
  drawParticules();
  glPopMatrix();
  
  SDL_GL_SwapBuffers();
}

/*
 * sdlFPS : compter le nombre de frames affiches. imprimer le
 * framerate toutes les 5 secondes.
 */
static inline void sdlFPS(void) {
  Uint32 t;
  static Uint32 t0 = 0, f = 0;
  f++;
  t = SDL_GetTicks();
  if(t - t0 > 5000) {
    fprintf(stderr, "sdlFPS:%6d\n",
	    (int)(1000.0 * f / (t - t0)));
    t0 = t;
    f  = 0;
  }
}

/*
 * Le cube ...
 */
static void cube(void) {
  float c[4] = {10, 10, 10, 0};
  glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, c);
/*   glDisable(GL_CULL_FACE); */
  /* Face arriere */
  glBegin(GL_LINE_LOOP);
  glVertex3f(-1,  1, -1);
  glVertex3f( 1,  1, -1);
  glVertex3f( 1, -1, -1);
  glVertex3f(-1, -1, -1);
  glEnd();
  /* Face gauche */
  glBegin(GL_LINE_LOOP);
  glVertex3f(-1, -1,  1);
  glVertex3f(-1,  1,  1);
  glVertex3f(-1,  1, -1);
  glVertex3f(-1, -1, -1);
  glEnd();
  /* Face droite */
  glBegin(GL_LINE_LOOP);
  glVertex3f( 1, -1,  1);
  glVertex3f( 1, -1, -1);
  glVertex3f( 1,  1, -1);
  glVertex3f( 1,  1,  1);
  glEnd();
  /* Face avant */
  glBegin(GL_LINE_LOOP);
  glVertex3f(-1,  1,  1);
  glVertex3f(-1, -1,  1);
  glVertex3f( 1, -1,  1);
  glVertex3f( 1,  1,  1);
  glEnd();
  /* Face haut */
  glBegin(GL_LINE_LOOP);
  glVertex3f(-1,  1,  1);
  glVertex3f( 1,  1,  1);
  glVertex3f( 1,  1, -1);
  glVertex3f(-1,  1, -1);
  glEnd();
  /* Face bas */
  glBegin(GL_LINE_LOOP);
  glVertex3f(-1, -1,  1);
  glVertex3f(-1, -1, -1);
  glVertex3f( 1, -1, -1);
  glVertex3f( 1, -1,  1);
  glEnd();
}
